import re
import unicodedata

# https://gist.github.com/ryanmcgrath/982242
# UNICODE RANGE : DESCRIPTION
# 3000-303F : punctuation
# 3040-309F : hiragana
# 30A0-30FF : katakana
# FF00-FFEF : Full-width roman + half-width katakana
# 4E00-9FAF : Common and uncommon kanji
#
# Non-Japanese punctuation/formatting characters commonly used in Japanese text
# 2605-2606 : Stars
# 2190-2195 : Arrows
# u203B     : Weird asterisk thing
import discord

JA_REGEX_STR = (r'[\u3000-\u303F]|[\u3040-\u309F]|[\u30A0-\u30FF]|[\uFF00-\uFFEF]'
                r'|[\u4E00-\u9FAF]|[\u2605-\u2606]|[\u2190-\u2195]|\u203B')
JA_REGEX = re.compile(JA_REGEX_STR)


def contains_ja(txt: str) -> bool:
    return bool(JA_REGEX.search(txt))


def normalize_server_name(server: str) -> str:
    server = server.upper()
    return 'NA' if server == 'US' else server


def is_valid_image_url(url: str) -> bool:
    url = url.lower()
    return url.startswith('http') and (url.endswith('.png') or url.endswith('.jpg'))


def extract_image_url(m: discord.Message) -> str:
    if is_valid_image_url(m.content):
        return m.content
    if m.attachments and len(m.attachments) and is_valid_image_url(m.attachments[0].url):
        return m.attachments[0].url
    return None


def rmdiacritics(string: str) -> str:
    """ Return the base character of char, by "removing" any
    diacritics like accents or curls and strokes and the like.
    """
    output = ''
    for c in string:
        try:
            desc = unicodedata.name(c)
            cutoff = desc.find(' WITH ')
            if cutoff != -1:
                desc = desc[:cutoff]
            output += unicodedata.lookup(desc)
        except (KeyError, ValueError):
            output += c
    return re.sub("[\u201c\u201d]", '"', output)


def clean_global_mentions(content: str) -> str:
    """Wipes out mentions to @everyone and @here."""
    return re.sub(r'(@)(\w)', '\\g<1>\u200b\\g<2>', content)


def strip_right_multiline(txt: str) -> str:
    """Useful for prettytable output where there are a lot of right spaces."""
    return '\n'.join([x.strip() for x in txt.splitlines()])



from discordmenu.embed.components import EmbedFooter
from discordmenu.embed.view_state import ViewState
from discordmenu.intra_message_state import IntraMessageState

from ..tsubaki import CLOUDFRONT_URL

TSUBAKI_FLOWER_ICON_URL = CLOUDFRONT_URL + '/tsubaki/tsubakiflower.png'


def embed_footer():
    return EmbedFooter('Requester may click the reactions below to switch tabs')


def embed_footer_with_state(state: ViewState, *, image_url=None, text=None):
    if image_url is None:
        image_url = TSUBAKI_FLOWER_ICON_URL
    if text is None:
        text = 'Requester may click the reactions below to switch tabs'

    url = IntraMessageState.serialize(image_url, state.serialize())
    return EmbedFooter(
        text,
        icon_url=url)



from redbot.core import data_manager

from .json_utils import *


class CogSettings(object):
    SETTINGS_FILE_NAME = "legacy_settings.json"

    def __init__(self, cog_name, bot=None):
        self.folder = str(data_manager.cog_data_path(raw_name=cog_name))
        self.file_path = os.path.join(self.folder, CogSettings.SETTINGS_FILE_NAME)

        self.bot = bot

        self.check_folder()

        self.default_settings = self.make_default_settings()
        if not os.path.isfile(self.file_path):
            logger.warning("CogSettings config for {} not found.  Creating default...".format(self.file_path))
            self.bot_settings = self.default_settings
            self.save_settings()
        else:
            current = self.intify(read_json_file(self.file_path))
            updated = False
            for key in self.default_settings.keys():
                if key not in current.keys():
                    current[key] = self.default_settings[key]
                    updated = True

            self.bot_settings = current
            if updated:
                self.save_settings()

    def check_folder(self):
        if not os.path.exists(self.folder):
            logger.info("Creating {}".format(self.folder))
            os.makedirs(self.folder)

    def save_settings(self):
        write_json_file(self.file_path, self.bot_settings)

    def make_default_settings(self):
        return {}

    @classmethod
    def intify(cls, key):
        if isinstance(key, dict):
            return {cls.intify(k): cls.intify(v) for k, v in key.items()}
        elif isinstance(key, (list, tuple)):
            return [cls.intify(x) for x in key]
        elif isinstance(key, str) and key.isdigit():
            return int(key)
        elif isinstance(key, str) and key.replace('.', '', 1).isdigit():
            return float(key)
        else:
            return key



from enum import Enum

from discord.ext.commands import BadArgument


class Server(Enum):
    COMBINED = "COMBINED"
    JP = "JP"
    NA = "NA"
    KR = "KR"

    @classmethod
    async def convert(cls, ctx, argument):
        if argument.upper() == ('JP', 'JA'):
            return Server.JP
        elif argument.upper() in ('NA', 'EN', 'US'):
            return Server.NA
        elif argument.upper() in ('KR', 'KO'):
            return Server.KO
        else:
            raise BadArgument(f"Unable to convert server `{argument}`")


class StarterGroup(Enum):
    RED = 0
    BLUE = 1
    GREEN = 2
    LIGHT = 3
    DARK = 4

    @classmethod
    async def convert(cls, ctx, argument):
        if 'RED'.startswith(argument.upper()):
            return StarterGroup.RED
        elif 'BLUE'.startswith(argument.upper()):
            return StarterGroup.BLUE
        elif 'GREEN'.startswith(argument.upper()):
            return StarterGroup.GREEN
        else:
            raise BadArgument(f"Unable to convert group `{argument}`")


class EvoToFocus(Enum):
    newest = 0
    naprio = 1


class AltEvoSort(Enum):
    dfs = 0
    numerical = 1


class ChildMenuType(Enum):
    IdMenu = 0
    NaDiffMenu = 1
    AwakeningList = 2


class LsMultiplier(Enum):
    lsdouble = 0
    lssingle = 1


class CardPlusModifier(Enum):
    plus0 = 0
    plus297 = 1


class EvoGrouping(Enum):
    splitevos = 0
    groupevos = 1



import re
from enum import Enum
from typing import Any, Dict

from tsutils.enums import AltEvoSort, CardPlusModifier, ChildMenuType, EvoGrouping, EvoToFocus, LsMultiplier, Server

SETTINGS_REGEX = re.compile(r'(?:--|—)(\w+)(?::{(.+?)})?')


class QuerySettings:
    SERIALIZED_VALUES = ['server', 'evosort', 'child_menu_type', 'lsmultiplier', 'cardplus', 'evogrouping']
    NAMES_TO_ENUMS = {
        'na_prio': EvoToFocus,
        'server': Server,
        'evosort': AltEvoSort,
        'child_menu_type': ChildMenuType,
        'lsmultiplier': LsMultiplier,
        'cardplus': CardPlusModifier,
        'evogrouping': EvoGrouping,
    }
    ENUMS_TO_NAMES = {v: k for k, v in NAMES_TO_ENUMS.items()}
    SETTINGS_TO_ENUMS = {
        "na": Server.NA,
        "allservers": Server.COMBINED,
        "dfs": AltEvoSort.dfs,
        "numerical": AltEvoSort.numerical,
        "nadiff": ChildMenuType.NaDiffMenu,
        "nadiffs": ChildMenuType.NaDiffMenu,
        "awakening": ChildMenuType.AwakeningList,
        "awakenings": ChildMenuType.AwakeningList,
        'lsdouble': LsMultiplier.lsdouble,
        'lssingle': LsMultiplier.lssingle,
        'lss': LsMultiplier.lssingle,
        'plus0': CardPlusModifier.plus0,
        'plus297': CardPlusModifier.plus297,
        'splitevos': EvoGrouping.splitevos,
        'groupevos': EvoGrouping.groupevos,
    }

    def __init__(self,
                 na_prio: EvoToFocus = EvoToFocus.naprio,
                 server: Server = Server.COMBINED,
                 evosort: AltEvoSort = AltEvoSort.dfs,
                 child_menu_type: ChildMenuType = ChildMenuType.IdMenu,
                 lsmultiplier: LsMultiplier = LsMultiplier.lsdouble,
                 cardplus: CardPlusModifier = CardPlusModifier.plus297,
                 evogrouping: EvoGrouping = EvoGrouping.splitevos,
                 ):
        self.na_prio = na_prio
        self.server = server
        self.evosort = evosort
        self.child_menu_type = child_menu_type
        self.lsmultiplier = lsmultiplier
        self.cardplus = cardplus
        self.evogrouping = evogrouping

    @classmethod
    def extract(cls, fm_flags: Dict[str, Any], query: str) -> "QuerySettings":
        fm_flags = fm_flags.copy()
        for key, value in fm_flags.items():
            if not isinstance(value, Enum):
                fm_flags[key] = QuerySettings.NAMES_TO_ENUMS[key](value)  # noqa

        for setting, data in re.findall(SETTINGS_REGEX, query.lower()):
            if setting not in cls.SETTINGS_TO_ENUMS:
                continue
            value = cls.SETTINGS_TO_ENUMS[setting]
            key = cls.ENUMS_TO_NAMES[type(value)]
            fm_flags[key] = value

        return QuerySettings(**fm_flags)

    def serialize(self: "QuerySettings") -> Dict[str, Any]:
        return {v: getattr(self, v).value for v in self.SERIALIZED_VALUES}

    @staticmethod
    def deserialize(data: Dict[str, Any]) -> "QuerySettings":
        enumdata = {}
        for key, value in data.items():
            enumdata[key] = QuerySettings.NAMES_TO_ENUMS[key](value)  # noqa

        return QuerySettings(**enumdata)



import asyncio
import io

from tsutils.enums import Server
from tsutils.errors import IndexNotLoaded


class aobject(object):
    """Base class to allow for asynchronous __init__"""

    # noinspection PyTypeChecker
    # noinspection PyMethodParameters
    async def __anew__(cls, *args, **kwargs):
        self = super().__new__(cls)
        await self.__init__(*args, **kwargs)
        return self

    async def __ainit__(self, *args, **kwargs):
        pass

    __new__ = __anew__
    __init__ = __ainit__


class CtxIO(io.IOBase):
    def __init__(self, ctx):
        self.ctx = ctx
        super(CtxIO, self).__init__()

    def read(self):
        raise io.UnsupportedOperation("read")

    def write(self, data):
        asyncio.ensure_future(self.ctx.send(data))


class DummyObject(dict):
    def __init__(self, item=None, **kwargs):
        if item is None:
            item = {}
        item.update(kwargs)
        super().__init__(item)
        self.__dict__ = item

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")


class IndexDict(dict):
    def __getattribute__(self, name):
        try:
            super().__getattribute__(name)
        except KeyError:
            if isinstance(name, Server):
                raise IndexNotLoaded("Index not yet loaded")
            raise



import asyncio
import re
from typing import Literal, Optional, List

import discord

from .emoji import SendableEmoji, YES_EMOJI, NO_EMOJI


async def send_repeated_consecutive_messages(ctx, message: str) -> discord.Message:
    """Edit the last message to include the string `x2` or more if would otherwise be repeated"""
    lmessage = await ctx.history().__anext__()
    fullmatch = re.escape(message) + r"(?: x(\d+))?"
    match = re.match(fullmatch, lmessage.content)
    if match and lmessage.author == ctx.bot.user:
        n = match.group(1) or "1"
        await lmessage.edit(content=message + " x" + str(int(n) + 1))
        return lmessage
    else:
        return await ctx.send(message)


async def get_user_confirmation(ctx, text: str,
                                yes_emoji: SendableEmoji = YES_EMOJI, no_emoji: SendableEmoji = NO_EMOJI,
                                timeout: int = 10) -> Literal[True, False, None]:
    msg = await ctx.send(text)
    asyncio.create_task(msg.add_reaction(yes_emoji))
    asyncio.create_task(msg.add_reaction(no_emoji))

    def check(reaction, user):
        return (str(reaction.emoji) in [yes_emoji, no_emoji]
                and user.id == ctx.author.id
                and reaction.message.id == msg.id)

    ret = False
    try:
        r, u = await ctx.bot.wait_for('reaction_add', check=check, timeout=timeout)
        if r.emoji == yes_emoji:
            ret = True
    except asyncio.TimeoutError:
        ret = None

    await msg.delete()
    return ret


async def get_user_reaction(ctx, text: str, *emoji: SendableEmoji, timeout: int = 10) -> Optional[SendableEmoji]:
    msg = await ctx.send(text)

    async def addreactions():
        for e in emoji:
            try:
                asyncio.create_task(msg.add_reaction(e))
            except (discord.Forbidden, discord.NotFound):
                pass

    asyncio.create_task(addreactions())

    def check(reaction, user):
        return (str(reaction.emoji) in emoji
                and user.id == ctx.author.id
                and reaction.message.id == msg.id)

    try:
        r, u = await ctx.bot.wait_for('reaction_add', check=check, timeout=timeout)
        ret = r.emoji
    except asyncio.TimeoutError:
        ret = None

    await msg.delete()
    return ret


async def await_and_remove(bot, react_msg: discord.Message, listen_user: discord.User,
                           delete_msgs: Optional[List[discord.Message]] = None,
                           emoji: SendableEmoji = NO_EMOJI, timeout: int = 15) -> None:
    """Remove a message when recieving an emoji reaction"""
    try:
        await react_msg.add_reaction(emoji)
    except Exception as e:
        # failed to add reaction, ignore
        return

    def check(payload):
        return str(payload.emoji.name) == emoji and \
               payload.user_id == listen_user.id and \
               payload.message_id == react_msg.id

    try:
        p = await bot.wait_for('add_reaction', check=check, timeout=timeout)
    except asyncio.TimeoutError:
        # Expected after {timeout} seconds
        p = None

    if p is None:
        try:
            await react_msg.remove_reaction(emoji, react_msg.guild.me)
        except Exception as e:
            # failed to remove reaction, ignore
            return
    else:
        msgs = delete_msgs or [react_msg]
        for m in msgs:
            await m.delete_message()


class StatusManager:
    """An asynchronous context manager to temporary modify the bot's status"""

    def __init__(self, bot, status: discord.Status = discord.Status.dnd):
        self.bot = bot
        self.newstatus = status
        self.oldstatus = None

    async def __aenter__(self) -> None:
        if not self.bot.guilds:
            return
        self.oldstatus = self.bot.guilds[0].me.status
        activity = self.bot.guilds[0].me.activity
        await self.bot.change_presence(activity=activity, status=self.newstatus)

    async def __aexit__(self, *args):
        if self.oldstatus is None \
                or not self.bot.guilds \
                or self.bot.guilds[0].me.status != self.newstatus:
            return
        activity = self.bot.guilds[0].me.activity
        await self.bot.change_presence(activity=activity, status=self.oldstatus)


def confirmation_message(msg: str) -> str:
    return f"{YES_EMOJI} {msg}"


async def send_confirmation_message(ctx, msg: str) -> discord.Message:
    return await ctx.send(confirmation_message(msg))


def cancellation_message(msg: str) -> str:
    return f"{NO_EMOJI} {msg}"


async def send_cancellation_message(ctx, msg: str) -> discord.Message:
    return await ctx.send(cancellation_message(msg))



emoji_buttons = {
    'home': '\N{HOUSE BUILDING}',
    'reset': '\N{WHITE MEDIUM STAR}',
}


class MenuPanes:
    INITIAL_EMOJI = emoji_buttons['home'],
    DATA = {}
    HIDDEN_EMOJIS = []

    @classmethod
    def emoji_names(cls):
        """Return all emoji names. If an emoji is specified as a tuple with fallback(s), returns a tuple."""
        return [k for k, v in cls.DATA.items() if k not in cls.HIDDEN_EMOJIS]

    @classmethod
    def all_emoji_names(cls):
        """
        Return all valid emoji names, including fallbacks.
        In particular, the length of the list returned by this method is not guaranteed to be
        the same as the length of the list returned by `emoji_names`, and so this should only
        be used for validity checking.
        """
        ret = []
        for k, v in cls.DATA.items():
            if k in cls.HIDDEN_EMOJIS:
                continue
            if isinstance(k, str):
                ret.append(k)
            elif isinstance(k, tuple):
                # case when a fallback is provided
                ret += k
        return ret

    @classmethod
    def transitions(cls):
        ret = {}
        for k, v in cls.DATA.items():
            if v[0] is None:
                continue
            if isinstance(k, str):
                ret[k] = v[0]
            elif isinstance(k, tuple):
                # case when a fallback is provided
                ret.update({_: v[0] for _ in k})
        return ret

    @classmethod
    def pane_types(cls):
        return {v[1]: v[0] for k, v in cls.DATA.items() if v[1] and v[1] not in cls.HIDDEN_EMOJIS}

    @classmethod
    def respond_to_emoji_with_parent(cls, emoji: str):
        """Only defined for menus that support children"""
        if cls.DATA.get(emoji) is None:
            return None
        return cls.DATA[emoji][0] is not None

    @classmethod
    def respond_to_emoji_with_child(cls, emoji: str):
        """Only defined for menus that support children"""
        if cls.DATA.get(emoji) is None:
            return None
        return cls.DATA[emoji][2] is not None

    @classmethod
    def get_child_data_func(cls, emoji: str):
        """Only defined for menus that support children"""
        if cls.DATA.get(emoji) is None:
            return None
        return cls.DATA[emoji][2]



import json
import logging
import os
import time

import aiohttp
import backoff

logger = logging.getLogger('tsutils')


def should_download(file_path, expiry_secs):
    if not os.path.exists(file_path):
        logger.debug("file does not exist, downloading " + file_path)
        return True

    ftime = os.path.getmtime(file_path)
    file_age = time.time() - ftime

    if file_age > expiry_secs:
        logger.debug("file {} too old, download it".format(file_path))
        return True
    else:
        return False


def write_json_file(file_path, js_data):
    with open(file_path, "w") as f:
        json.dump(js_data, f, indent=4)


def read_json_file(file_path):
    with open(file_path) as f:
        return json.load(f)


def safe_read_json(file_path):
    """This returns an empty dict rather than raising an error if the file contains invalid json"""
    try:
        return read_json_file(file_path)
    except (json.JSONDecodeError, FileNotFoundError):
        logger.error('failed to read {} got exception'.format(file_path), exc_info=True)
    return {}


def validate_json(fp):
    try:
        json.load(open(fp))
        return True
    except json.JSONDecodeError:
        return False


@backoff.on_exception(backoff.expo, aiohttp.ClientError, max_time=60)
@backoff.on_exception(backoff.expo, aiohttp.ServerDisconnectedError, max_time=60)
async def async_cached_dadguide_request(file_path, file_url, expiry_secs):
    if should_download(file_path, expiry_secs):
        async with aiohttp.ClientSession() as session:
            async with session.get(file_url) as resp:
                assert resp.status == 200
                with open(file_path, 'wb') as f:
                    f.write(await resp.read())


def write_plain_file(file_path, text_data):
    with open(file_path, "w", encoding='utf-8') as f:
        f.write(text_data)


def read_plain_file(file_path):
    with open(file_path, "r", encoding='utf-8') as f:
        return f.read()


async def async_plain_request(file_url):
    async with aiohttp.ClientSession() as session:
        async with session.get(file_url) as resp:
            return await resp.text()


async def async_cached_plain_request(file_path, file_url, expiry_secs):
    if should_download(file_path, expiry_secs):
        resp = await async_plain_request(file_url)
        write_plain_file(file_path, resp)
    return read_plain_file(file_path)



import asyncio
import inspect

import discord


class EmojiUpdater(object):
    # a pass-through base class that does nothing to the emoji dictionary
    # or to the selected emoji
    def __init__(self, emoji_to_embed, **kwargs):
        self.emoji_dict = emoji_to_embed
        self.selected_emoji = None

    async def on_update(self, ctx, selected_emoji):
        self.selected_emoji = selected_emoji
        return True


class Menu():
    """Menu by https://github.com/Awoonar/Dusty-Cogs/blob/master/menu/menu.py"""

    def __init__(self, bot):
        self.bot = bot

        # Feel free to override this in your cog if you need to
        self.emoji = {
            0: "0⃣",
            1: "1⃣",
            2: "2⃣",
            3: "3⃣",
            4: "4⃣",
            5: "5⃣",
            6: "6⃣",
            7: "7⃣",
            8: "8⃣",
            9: "9⃣",
            10: "🔟",
            "next": "➡",
            "back": "⬅",
            "yes": "✅",
            "no": "❌",
        }

    # for use as an action
    async def reaction_delete_message(self, bot, ctx, message):
        await message.delete()

    async def custom_menu(self, ctx, emoji_to_message, selected_emoji, **kwargs):
        """Creates and manages a new menu
        Required arguments:
            Type:
                1- number menu
                2- confirmation menu
                3- info menu (basically menu pagination)
                4- custom menu. If selected, choices must be a list of tuples.
            Messages:
                Strings or embeds to use for the menu.
                Pass as a list for number menu
        Optional arguments:
            page (Defaults to 0):
                The message in messages that will be displayed
            timeout (Defaults to 15):
                The number of seconds until the menu automatically expires
            check (Defaults to default_check):
                The same check that wait_for_reaction takes
            is_open (Defaults to False):
                Whether or not the menu can take input from any user
            emoji (Defaults to self.emoji):
                A dictionary containing emoji to use for the menu.
                If you pass this, use the same naming scheme as self.emoji
            message (Defaults to None):
                The discord.Message to edit if present
            """
        return await self._custom_menu(ctx, emoji_to_message, selected_emoji, **kwargs)

    async def show_menu(self,
                        ctx,
                        message,
                        new_message_content):
        if message:
            if isinstance(new_message_content, discord.Embed):
                return await message.edit(embed=new_message_content)
            else:
                return await message.edit(content=new_message_content)
        else:
            if isinstance(new_message_content, discord.Embed):
                return await ctx.send(embed=new_message_content)
            else:
                return await ctx.send(new_message_content)

    async def _custom_menu(self, ctx, emoji_to_message, selected_emoji,
                           allowed_action=True, **kwargs):
        timeout = kwargs.get('timeout', 15)
        message = kwargs.get('message', None)

        reactions_required = not message
        new_message_content = emoji_to_message.emoji_dict[selected_emoji]
        if allowed_action:
            if not message:
                message = await self.show_menu(ctx, message, new_message_content)
            else:
                await self.show_menu(ctx, message, new_message_content)

        if reactions_required:
            async def addreactions():
                for e in emoji_to_message.emoji_dict:
                    try:
                        await message.add_reaction(e)
                    except discord.Forbidden:
                        pass

            asyncio.create_task(addreactions())

        def check(payload):
            def default_check(pl):
                return pl.user_id != self.bot.user.id and not (pl.guild_id and pl.member.bot)

            return (kwargs.get('check', default_check)(payload) and
                    str(payload.emoji.name) in list(emoji_to_message.emoji_dict.keys()) and
                    payload.user_id == ctx.author.id and
                    payload.message_id == message.id)

        if not message:
            raise ValueError(str((message, ctx)))

        try:
            p = await self.bot.wait_for('raw_reaction_add', check=check, timeout=timeout)
        except asyncio.TimeoutError:
            p = None

        if p is None:
            try:
                await message.clear_reactions()
            except discord.Forbidden:
                pass
            return message, new_message_content

        react_emoji = p.emoji.name
        react_action = emoji_to_message.emoji_dict[p.emoji.name]

        if inspect.iscoroutinefunction(react_action):
            message = await react_action(self.bot, ctx, message)
        elif inspect.isfunction(react_action):
            message = react_action(ctx, message)

        # user function killed message, quit
        if not message:
            return None, None

        try:
            if not isinstance(message.channel, discord.DMChannel):
                await message.remove_reaction(react_emoji, p.member)
        except discord.Forbidden:
            pass

        # update the emoji mapping however we need to, or just pass through and do nothing

        allowed_action = await emoji_to_message.on_update(ctx, react_emoji)
        return await self._custom_menu(
            ctx, emoji_to_message, emoji_to_message.selected_emoji,
            timeout=timeout,
            check=check,
            message=message,
            allowed_action=allowed_action)



from .cog_settings import *
from .emoji import *
from .formatting import *
from .helper_classes import *
from .helper_functions import *
from .json_utils import *
from .old_menu import *
from .menu.footers import *
from .pad import *
from .tsubaki import *
from .user_interaction import *



import asyncio
import discord.ext
import errno
import os
import signal
from functools import wraps


def timeout_after(seconds=10, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.setitimer(signal.ITIMER_REAL, seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wraps(func)(wrapper)

    return decorator


def corowrap(coro, loop):
    def func(*args, **kwargs):
        fut = asyncio.run_coroutine_threadsafe(coro, loop)
        try:
            fut.result()
        except Exception:
            pass

    return func


def fawait(coro, loop):
    fut = asyncio.run_coroutine_threadsafe(coro, loop)
    try:
        fut.result()
    except Exception:
        pass


async def repeating_timer(seconds, condition=None, start_immediately=True):
    if condition is None:
        def condition():
            return None
    if start_immediately:
        yield
    while condition():
        await asyncio.sleep(seconds)
        yield


def deepget(mapping, keys, default):
    o = mapping
    for key in keys:
        try:
            o = o[key]
        except KeyError:
            return default
    return o


def make_non_gatekeeping_check(condition, failmessage):
    """Creates a check decorator that doesn't stop the a user from seeing a command.

    It only shows a "You don't have permission to use this" string (specified by failmessage)
    if an invalid user tries to use it.
    """

    def non_gatekeep_check(**kwargs):
        def decorator(command):
            @command.before_invoke
            async def hook(instance, ctx):
                if not condition(ctx, **kwargs):
                    await ctx.send(failmessage.format(ctx))
                    raise discord.ext.commands.CheckFailure()

            return command

        return decorator

    return non_gatekeep_check



from redbot.core.commands import UserFeedbackCheckFailure


class IndexNotLoaded(KeyError):
    pass


class ClientInlineTextException(UserFeedbackCheckFailure):  # TODO: Find another way to do this
    """An error that sends its message to the user in the Discord client instead of
    failing a command and printing a message to the terminal/logger

    This error has nothing to do with checks, and only inherits from UserFeedbackCheckFailure
    because the Red bot instance automatically sends its message to the current
    channel when raised.
    """
    pass


class NoAPIKeyException(ClientInlineTextException):
    def __init__(self, fix_command=None, *args):
        if fix_command is not None:
            super().__init__(f"API key not found.  Set it with `{fix_command}`", *args)
        else:
            super().__init__(None, *args)


class BadAPIKeyException(ClientInlineTextException):
    def __init__(self, fix_command=None, *args):
        if fix_command is not None:
            super().__init__(f"Invalid API key.  Fix it with `{fix_command}`", *args)
        else:
            super().__init__(None, *args)



from redbot.core import commands

from .helper_functions import make_non_gatekeeping_check

CLOUDFRONT_URL = "https://d30r6ivozz8w2a.cloudfront.net"


def auth_check(perm, default=False):
    def check(ctx):
        if ctx.author.id in ctx.bot.owner_ids:
            return True
        authcog = ctx.bot.get_cog("GlobalAdmin")
        if not authcog:
            return default
        return authcog.settings.get_perm(ctx.author.id, perm, default=default)

    return commands.check(check)


def _is_donor(ctx, only_patron=False):
    if ctx.author.id in ctx.bot.owner_ids:
        return True
    donationcog = ctx.bot.get_cog("Donations")
    if not donationcog:
        return False
    return donationcog.is_donor(ctx, only_patron)


is_donor = make_non_gatekeeping_check(
    _is_donor,
    ("Sorry, but this is a Donor Only command.  Learn "
     "more about donation via `{0.prefix}donate`")
)



import re
from typing import TYPE_CHECKING
from discord.ext.commands import Converter, BadArgument
import discord

if TYPE_CHECKING:
    CaseInsensitiveRole = discord.Role
else:
    class CaseInsensitiveRole(Converter):
        async def convert(self, ctx, argument) -> discord.Role:
            match = re.match(r'(?:<@&)?(\d{15,20})>?$', argument)
            if match and ctx.guild.get_role(int(match.group(1))):
                return ctx.guild.get_role(int(match.group(1)))

            roles = {role for role in ctx.guild.roles if role.name == argument} or \
                    {role for role in ctx.guild.roles if role.name.lower() == argument.lower()}

            if len(roles) == 1:
                return roles.pop()
            if len(roles) == 0:
                raise BadArgument(f"No role matched {argument}")

            highest_prio = max(roles, key=lambda r: r.position)
            await ctx.send(f"Warning: More than one candidate found: {', '.join(r.mention for r in roles)}"
                           f" for input `{argument}`. We matched {highest_prio.mention}.",
                           allowed_mentions=discord.AllowedMentions(everyone=False, roles=False))
            return highest_prio



import re
from typing import Sequence, Union

import discord

SendableEmoji = Union[str, discord.Emoji]
ReceivableEmoji = Union[SendableEmoji, discord.PartialEmoji]

YES_EMOJI = '\N{WHITE HEAVY CHECK MARK}'
NO_EMOJI = '\N{CROSS MARK}'

def char_to_emoji(c: str) -> SendableEmoji:
    """Casts individual characters & the integer 10 as emojis"""
    c = str(c).lower()
    if '0' <= c <= '9':
        names = {
            '0': '0\N{COMBINING ENCLOSING KEYCAP}',
            '1': '1\N{COMBINING ENCLOSING KEYCAP}',
            '2': '2\N{COMBINING ENCLOSING KEYCAP}',
            '3': '3\N{COMBINING ENCLOSING KEYCAP}',
            '4': '4\N{COMBINING ENCLOSING KEYCAP}',
            '5': '5\N{COMBINING ENCLOSING KEYCAP}',
            '6': '6\N{COMBINING ENCLOSING KEYCAP}',
            '7': '7\N{COMBINING ENCLOSING KEYCAP}',
            '8': '8\N{COMBINING ENCLOSING KEYCAP}',
            '9': '9\N{COMBINING ENCLOSING KEYCAP}',
            '10': '\N{KEYCAP TEN}'
        }
        return names[c]
    if c < 'a' or c > 'z':
        return c

    base = ord('\N{REGIONAL INDICATOR SYMBOL LETTER A}')
    adjustment = ord(c) - ord('a')
    return chr(base + adjustment)


def fix_emojis_for_server(emoji_list: Sequence[discord.Emoji], msg_text: str) -> str:
    """Finds 'emoji-looking' substrings in msg_text and corrects them.

    If msg_text has something like '<:emoji_1_derp:13242342343>' and the server
    contains an emoji named :emoji_2_derp: then it will be swapped out in
    the message.

    This corrects an issue where a padglobal alias is created in one server
    with an emoji, but it has a slightly different name in another server.
    """
    # Find all emoji-looking things in the message
    matches = re.findall(r'<:[0-9a-z_]+:\d{18}>', msg_text, re.IGNORECASE)
    if not matches:
        return msg_text

    # For each unique looking emoji thing
    for m in set(matches):
        # Create a regex for that emoji replacing the digit
        m_re = re.sub(r'\d', r'&', m).rstrip("~")
        for em in emoji_list:
            # If the current emoji matches the regex, force a replacement
            emoji_code = str(em)
            if re.match(m_re, emoji_code, re.IGNORECASE):
                msg_text = re.sub(m_re, emoji_code, msg_text, flags=re.IGNORECASE)
                break
    return msg_text


def replace_emoji_names_with_code(emoji_list: Sequence[discord.Emoji], msg_text: str) -> str:
    """Finds emoji-name substrings in msg_text and corrects them.

    If msg_text has something like ':emoji_1_derp:' and emoji_list contains
    an emoji named 'emoji_1_derp' then the value will replaced with the full
    emoji id.

    This allows a padglobal admin without nitro to create entries with emojis
    from other servers.
    """
    # First strip down actual emojis to just the names
    msg_text = re.sub(r'<(:[0-9a-z_]+:)\d{18}>', r'\1', msg_text, flags=re.IGNORECASE)

    # Find all emoji-looking things in the message
    matches = re.findall(r':[0-9a-z_]+:', msg_text, re.IGNORECASE)
    if not matches:
        return msg_text

    # For each unique looking emoji thing
    finished_names = []
    for m in set(matches):
        emoji_name = m.strip(':')
        for e in emoji_list:
            if e.name == emoji_name and e.name not in finished_names:
                finished_names.append(e.name)
                msg_text = msg_text.replace(m, str(e))
    return msg_text



import pytz

# TZ used for PAD NA
# NA_TZ_OBJ = pytz.timezone('America/Los_Angeles')
NA_TZ_OBJ = pytz.timezone('US/Pacific')

# TZ used for PAD JP
JP_TZ_OBJ = pytz.timezone('Asia/Tokyo')

# This was overwritten by voltron. PDX opted to copy it +10,000 ids away
CROWS_1 = {x: x + 10000 for x in range(2601, 2635 + 1)}
# This isn't overwritten but PDX adjusted anyway
CROWS_2 = {x: x + 10000 for x in range(3460, 3481 + 1)}

PDX_JP_ADJUSTMENTS = {}
PDX_JP_ADJUSTMENTS.update(CROWS_1)
PDX_JP_ADJUSTMENTS.update(CROWS_2)


def get_pdx_id(m):
    pdx_id = m.monster_no_na
    if int(m.monster_id) == m.monster_no_jp:
        pdx_id = PDX_JP_ADJUSTMENTS.get(pdx_id, pdx_id)
    return pdx_id


def get_pdx_id_dadguide(m):
    pdx_id = m.monster_no_na
    if int(m.monster_id) == m.monster_no_jp:
        pdx_id = PDX_JP_ADJUSTMENTS.get(pdx_id, pdx_id)
    return pdx_id
